from functools import cache
import json
from os import path
from threading import Thread

from celery import bootsteps
from celery.concurrency import solo
from celery.utils.log import get_logger
from confluent_kafka.error import KafkaException
from fastavro.schema import parse_schema
from hop import auth, Stream
from hop.io import list_topics
from hop.models import AvroBlob, JSONBlob
from xdg.BaseDirectory import xdg_config_home

from .signals import kafka_record_consumed
from ..util import read_json

__all__ = ('Producer', 'Consumer')

log = get_logger(__name__)


@cache
def schema():
    # The order does not matter other than the Alert schema must be loaded last
    # because it references the other schema. All of the schema are saved in
    # named_schemas, but we only need to save a reference to the the Alert
    # schema to write the packet.
    # NOTE Specifying expand=True when calling parse_schema is okay when only
    # one schema contains references to other schema, in our case only the
    # alerts schema contains references to other schema. More complicated
    # relationships between schema though can lead to behavior that does not
    # conform to the avro spec, and a different method will need to be used to
    # load the schema. See https://github.com/fastavro/fastavro/issues/624 for
    # more info.
    named_schemas = {}
    for s in ['igwn.alerts.v1_0.ExternalCoincInfo.avsc',
              'igwn.alerts.v1_0.EventInfo.avsc',
              'igwn.alerts.v1_0.AlertType.avsc',
              'igwn.alerts.v1_0.Alert.avsc']:
        schema = parse_schema(read_json('igwn_gwalert_schema', s),
                              named_schemas, expand=True)

    return schema


def _load_hopauth_map():
    hop_auth = auth.load_auth()
    with open(path.join(xdg_config_home,
                        'gwcelery/kafka_credential_map.json'),
              'r') as fo:
        kafka_credential_map = json.load(fo)

    return hop_auth, kafka_credential_map


class AvroBlobWrapper(AvroBlob):

    def __init__(self, payload):
        return super().__init__([payload], schema())


class KafkaBase:

    def __init__(self, name, config, prefix):
        self.name = name
        self._config = config
        self._credential = self.get_auth(prefix)
        self._hop_stream = Stream(self._credential)

        # FIXME Drop get_payload_content method once
        # https://github.com/scimma/hop-client/pull/190 is merged
        if config['suffix'] == 'avro':
            self.serialization_model = AvroBlobWrapper
            self.get_payload_content = lambda payload: payload.content[0]
        elif config['suffix'] == 'json':
            self.serialization_model = JSONBlob
            self.get_payload_content = lambda payload: payload.content
        else:
            raise NotImplementedError(
                'Supported serialization method required for alert notices'
            )

    def get_auth(self, prefix):
        hop_auth, kafka_credential_map = _load_hopauth_map()

        # kafka_credential_map contains map between logical name for broker
        # topic and username
        username = kafka_credential_map.get(prefix, {}).get(self.name)
        if username is None:
            raise ValueError('Unable to find {} entry in kafka credential map '
                             'for {}'.format(prefix, self.name))

        # hop auth contains map between username and password/hostname
        target_auth = None
        for cred in hop_auth:
            if cred.username != username:
                continue
            target_auth = cred
            break
        else:
            raise ValueError('Unable to find entry in hop auth file for '
                             'username {}'.format(username))
        return target_auth


class KafkaListener(KafkaBase):

    def __init__(self, name, config):
        super().__init__(name, config, 'consumer')
        self._open_hop_stream = None
        # Don't kill worker if listener can't connect
        try:
            self._open_hop_stream = self._hop_stream.open(config['url'], 'r')
        except KafkaException:
            log.exception('Connection to %s failed', self._config["url"])
        except ValueError:
            # Hop client will return a ValueError if the topic doesn't exist on
            # the broker
            log.exception('Connection to %s failed', self._config["url"])

    def listen(self):
        for message in self._open_hop_stream:
            # Send signal
            kafka_record_consumed.send(
                None,
                name=self.name,
                record=self.get_payload_output(message)
            )


class KafkaWriter(KafkaBase):
    """Write Kafka topics and monitor health."""

    def __init__(self, name, config):
        super().__init__(name, config, 'producer')
        self._open_hop_stream = self._hop_stream.open(
            config['url'], 'w',
            message_max_bytes=1024 * 1024 * 2,
            compression_type='zstd')

        # Set up flag for failed delivery of messages
        self.kafka_delivery_failures = False

    def kafka_topic_up(self):
        '''Check for problems in broker and topic. Returns True is broker and
        topic appear to be up, returns False otherwise.'''
        kafka_url = self._config['url']
        _, _, broker, topic = kafka_url.split('/')
        try:
            topics = list_topics(kafka_url, auth=self._credential, timeout=5)
            if topics[topic].error is None:
                log.info(f'{kafka_url} appears to be functioning properly')
                return True
            else:
                log.error(f'{topic} at {broker} appears to be down')
                return False
        except KafkaException:
            log.error(f'{broker} appears to be down')
            return False

    def _delivery_cb(self, kafka_error, message):
        # FIXME Get rid of if-else logic once
        # https://github.com/scimma/hop-client/pull/190 is merged
        if self._config['suffix'] == 'avro':
            record = AvroBlob.deserialize(message.value()).content[0]
        else:
            record = JSONBlob.deserialize(message.value()).content
        kafka_url = self._config['url']
        if kafka_error is None:
            self.kafka_delivery_failures = False
        else:
            log.error(f'Received error code {kafka_error.error_code} '
                      f'({kafka_error.reason}) when delivering '
                      f'{record["superevent_id"]} '
                      f'{record["alert_type"]} alert to {kafka_url}')
            self.kafka_delivery_failures = True

    def write(self, payload):
        self._open_hop_stream.write(payload,
                                    delivery_callback=self._delivery_cb)
        self._open_hop_stream.flush()


class KafkaBootStep(bootsteps.ConsumerStep):
    """Generic boot step to limit us to appropriate kinds of workers.

    Only include this bootstep in workers that are configured to listen to the
    ``kafka`` queue.
    """

    def include_if(self, consumer):
        return 'kafka' in consumer.app.amqp.queues

    def create(self, consumer):
        if not isinstance(consumer.pool, solo.TaskPool):
            raise RuntimeError(
                'The Kafka broker only works with the "solo" task pool. '
                'Start the worker with "--queues=kafka --pool=solo".')


class Consumer(KafkaBootStep):
    """Run MOU Kafka consumers in background threads.
    """

    name = 'Kafka consumer'

    def start(self, consumer):
        log.info(f'Starting {self.name}, topics: ' +
                 ' '.join(config['url'] for config in
                          consumer.app.conf['kafka_consumer_config'].values()))
        self._listeners = {
            key: KafkaListener(key, config) for key, config in
            consumer.app.conf['kafka_consumer_config'].items()
        }
        self.threads = [
            Thread(target=s.listen, name=f'{key}_KafkaConsumerThread') for key,
            s in self._listeners.items() if s._open_hop_stream is not None
        ]
        for thread in self.threads:
            thread.start()

    def stop(self, consumer):
        log.info('Closing connection to topics: ' +
                 ' '.join(listener._config['url'] for listener in
                          self._listeners.values() if listener._open_hop_stream
                          is not None))
        for s in self._listeners.values():
            if s._open_hop_stream is not None:
                s._open_hop_stream.close()

        for thread in self.threads:
            thread.join()

    def info(self, consumer):
        return {
            'active_kafka_consumers': {listener.name for listener in
                                       self._listeners.values() if
                                       listener._open_hop_stream is not
                                       None}
        }


class Producer(KafkaBootStep):
    """Run the global Kafka producers in a background thread.

    Flags that document the health of the connections are made available
    :ref:`inspection <celery:worker-inspect>` with the ``gwcelery inspect
    stats`` command under the ``kafka_topic_up`` and
    ``kafka_delivery_failures`` keys.
    """

    name = 'Kafka producer'

    def start(self, consumer):
        log.info(f'Starting {self.name}, topics: ' +
                 ' '.join(config['url'] for config in
                          consumer.app.conf['kafka_alert_config'].values()))
        consumer.app.conf['kafka_streams'] = self._writers = {
            brokerhost: KafkaWriter(brokerhost, config) for brokerhost, config
            in consumer.app.conf['kafka_alert_config'].items()
        }

    def stop(self, consumer):
        log.info('Closing connection to topics: ' +
                 ' '.join(config['url'] for config in
                          consumer.app.conf['kafka_alert_config'].values()))
        for s in self._writers.values():
            s._open_hop_stream.close()

    def info(self, consumer):
        return {'kafka_topic_up': {
                    brokerhost: writer.kafka_topic_up() for brokerhost,
                    writer in self._writers.items()
                },
                'kafka_delivery_failures': {
                    brokerhost: writer.kafka_delivery_failures for
                    brokerhost, writer in self._writers.items()
                }}



"""Shortcuts for HTCondor commands to manage deployment of GWCelery on LIGO
Data Grid clusters.

These commands apply to the GWCelery instance that is
running in the current working directory.
"""
from importlib import resources
import json
import os
import shlex
import subprocess
import sys
import time

import click
import lxml.etree

from .. import data

with resources.path(data, 'gwcelery.sub') as p:
    SUBMIT_FILE = str(p)


@click.group(help=__doc__)
def condor():
    # Just register the condor command group;
    # no need to do anything here.
    pass


def get_constraints():
    return '-constraint', 'JobBatchName=={} && Iwd=={}'.format(
        json.dumps('gwcelery'),  # JSON string literal escape sequences
        json.dumps(os.getcwd())  # are a close match to HTCondor ClassAds.
    )


def run_exec(*args):
    print(' '.join(shlex.quote(arg) for arg in args))
    os.execvp(args[0], args)


def running():
    """Determine if GWCelery is already running under HTCondor."""
    status = subprocess.check_output(('condor_q', '-xml', *get_constraints()))
    classads = lxml.etree.fromstring(status)
    return classads.find('.//c') is not None


@condor.command()
@click.pass_context
def submit(ctx):
    """Submit all GWCelery jobs to HTCondor (if not already running)."""
    if running():
        print('error: GWCelery jobs are already running in this directory.\n'
              'First remove existing jobs with "gwcelery condor rm".\n'
              'To see the status of those jobs, run "gwcelery condor q".',
              file=sys.stderr)
        sys.exit(1)
    else:
        accounting_group = ctx.obj.app.conf['condor_accounting_group']
        run_exec('condor_submit',
                 'accounting_group={}'.format(accounting_group),
                 SUBMIT_FILE)


@condor.command()
@click.pass_context
def resubmit(ctx):
    """Remove any running GWCelery jobs and resubmit to HTCondor."""
    if running():
        subprocess.check_call(('condor_rm', *get_constraints()))
    timeout = 120
    start = time.monotonic()
    while time.monotonic() - start < timeout:
        if not running():
            break
        time.sleep(1)
    else:
        print('error: Could not stop all GWCelery jobs', file=sys.stderr)
        sys.exit(1)
    accounting_group = ctx.obj.app.conf['condor_accounting_group']
    run_exec('condor_submit', 'accounting_group={}'.format(accounting_group),
             SUBMIT_FILE)


@condor.command()
def rm():
    """Remove all GWCelery jobs."""
    run_exec('condor_rm', *get_constraints())


@condor.command()
def hold():
    """Put all GWCelery jobs on hold."""
    run_exec('condor_hold', *get_constraints())


@condor.command()
def release():
    """Release all GWCelery jobs from hold status."""
    run_exec('condor_release', *get_constraints())


@condor.command()
def q():
    """Show status of all GWCelery jobs."""
    run_exec('condor_q', '-nobatch', *get_constraints())



import json
import pytest
from importlib import resources
from unittest.mock import Mock
from ..tasks import rrt_utils
from gwcelery.tests import data


skymap_small = resources.read_binary(data, 'rrt_small_area.fits')
skymap_large = resources.read_binary(
    data, 'MS220722v_bayestar.multiorder.fits')

# checks
# 1. Burst far checks   ---> True
# 2. HasRemnant check   ---> True
# 3. p_BNS check        ---> True
# 4. p_NSBH check       ---> True
# 5. skymap check       ---> True
# 6. Simple BBH         ---> False
# 7. Burst but low far  ---> False
# 8. High p-terrestrial ---> False
# 9. Burst BBH low far ---> False
# 10. HIGH_PROFILE_applied-> False


def get_event(graceid):
    if graceid == 'G123':
        event = {
            'group': 'CBC',
            'pipeline': 'gstlal',
            'search': 'AllSky',
            'graceid': graceid,
            'far': 1e-9,
            'gpstime': 1234,
        }
    elif graceid == 'G456':
        event = {
            'group': 'Burst',
            'pipeline': 'cwb',
            'search': 'AllSky',
            'graceid': graceid,
            'far': 1e-10,
            'gpstime': 1234,
        }
    elif graceid == 'G789':
        event = {
            'group': 'Burst',
            'pipeline': 'cwb',
            'search': 'AllSky',
            'graceid': graceid,
            'far': 1e-8,
            'gpstime': 1234,
        }
    elif graceid == 'G246':
        event = {
            'group': 'Burst',
            'pipeline': 'cwb',
            'search': 'BBH',
            'graceid': graceid,
            'far': 1e-10,
            'gpstime': 1234,
        }
    return event


@pytest.mark.parametrize(
        'superevent,embright,pastro,skymap,result',
        [[{'labels': [], 'gw_events':['G123', 'G456']},
          {'HasNS': 0.0, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 0.7, "Terrestrial": 0.3},
          skymap_large, True],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 0.8, 'HasRemnant': 0.2, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 0.7, "Terrestrial": 0.3},
          skymap_large, True],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 0.9, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.8, "NSBH": 0.2, "BBH": 0.0, "Terrestrial": 0.0},
          skymap_large, True],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 0.9, 'HasRemnant': 0.2, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.9, "BBH": 0.0, "Terrestrial": 0.1},
          skymap_large, True],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 0.0, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 0.9, "Terrestrial": 0.1},
          skymap_small, True],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 0.0, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 1.0, "Terrestrial": 0.0},
          skymap_large, False],
         [{'labels': [], 'gw_events':['G123', 'G789']},
          {'HasNS': 0.0, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 1.0, "Terrestrial": 0.0},
          skymap_large, False],
         [{'labels': [], 'gw_events':['G123']},
          {'HasNS': 1.0, 'HasRemnant': 1.0, 'HasMassGap': 0.0},
          {"BNS": 0.2, "NSBH": 0.2, "BBH": 0.0, "Terrestrial": 0.6},
          skymap_large, False],
         [{'labels': [], 'gw_events':['G123', 'G246']},
          {'HasNS': 0.0, 'HasRemnant': 0.0, 'HasMassGap': 0.0},
          {"BNS": 0.0, "NSBH": 0.0, "BBH": 0.7, "Terrestrial": 0.3},
          skymap_large, False],
         [{'labels': ['HIGH_PROFILE'], 'gw_events':['G123', 'G246']},
          {'HasNS': 1.0, 'HasRemnant': 1.0, 'HasMassGap': 0.0},
          {"BNS": 1.0, "NSBH": 0.0, "BBH": 0.0, "Terrestrial": 0.0},
          skymap_small, False]
         ])
def test_high_profile(monkeypatch, superevent, embright,
                      pastro, skymap, result):
    embright = json.dumps(embright)
    pastro = json.dumps(pastro)
    superevent.update({'superevent_id': 'S123',
                       'preferred_event_data': get_event('G123')})
    mock_create_label = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.create_label.run',
                        mock_create_label)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event.run',
                        get_event)
    rrt_utils.check_high_profile(skymap, embright,
                                 pastro, superevent)
    if result is True:
        mock_create_label.assert_called_once_with(
            "HIGH_PROFILE", "S123"
        )
    elif result is False:
        mock_create_label.assert_not_called()
